#include <stdio.h>

#ifdef GCC
#define restrict __restrict__
#endif

#ifdef CLANG
#define restrict
#endif

#if defined(HAS_VECT8_EXT)

typedef float v8sf __attribute__ ((vector_size (8 * sizeof(float))));

static inline void add3_vec(float * restrict arg0, float * restrict arg1, float * restrict arg2, unsigned int vecsize)
{
  unsigned int i;
  v8sf v0, v1, v2;

  for (i=0; i < vecsize; i+=8)
    {
      v0 = *((v8sf * restrict) (arg0 + i));
      v1 = *((v8sf * restrict) (arg1 + i));

      v2 = v0 + v1;

      *((v8sf * restrict) (arg2 + i)) = v2;
    }
}

static inline void mul3_vec(float * restrict arg0, float * restrict arg1,  float * restrict arg2, unsigned int vecsize)
{
  unsigned int i;
  v8sf v0, v1, v2;

  for (i=0; i < vecsize; i+=8)
    {
      v0 = *((v8sf * restrict) (arg0 + i));
      v1 = *((v8sf * restrict) (arg1 + i));

      v2 = v0 * v1;

      *((v8sf * restrict) (arg2 + i)) = v2;
    }
}

#elif defined(HAS_VECT4_EXT)

typedef float v4sf __attribute__ ((vector_size (4 * sizeof(float))));

static inline void add3_vec(float * restrict arg0, float * restrict arg1, float * restrict arg2, unsigned int vecsize)
{
  unsigned int i;
  v4sf v0, v1, v2;

  for (i=0; i < vecsize; i+=4)
    {
      v0 = *((v4sf * restrict) (arg0 + i));
      v1 = *((v4sf * restrict) (arg1 + i));

      v2 = v0 + v1;

      *((v4sf * restrict) (arg2 + i)) = v2;
    }
}

static inline void mul3_vec(float * restrict arg0, float * restrict arg1,  float * restrict arg2, unsigned int vecsize)
{
  unsigned int i;
  v4sf v0, v1, v2;

  for (i=0; i < vecsize; i+=4)
    {
      v0 = *((v4sf * restrict) (arg0 + i));
      v1 = *((v4sf * restrict) (arg1 + i));

      v2 = v0 * v1;

      *((v4sf * restrict) (arg2 + i)) = v2;
    }
}

#else

static inline void add3_vec(float * restrict arg0, float * restrict arg1, float * restrict arg2, unsigned int vecsize)
{
  unsigned int i;

  for (i = 0; i < vecsize; i++)
    arg2[i] = arg0[i] + arg1[i];
}

static inline void mul3_vec(float * restrict arg0, float * restrict arg1,  float * restrict arg2, unsigned int vecsize)
{
  unsigned int i;

  for (i = 0; i < vecsize; i++)
    arg2[i] = arg0[i] * arg1[i];
}

#endif

#define VECT_SIZE 256
#define ITERATIONS 100000000

int main(int argc, const char *argv[])
{
  float va[VECT_SIZE] __attribute__ ((__aligned__(8 * sizeof(float)))) ;
  float vb[VECT_SIZE]  __attribute__ ((__aligned__(8 * sizeof(float)))) ;
  float vc[VECT_SIZE]  __attribute__ ((__aligned__(8 * sizeof(float)))) ;
  float temp[VECT_SIZE]  __attribute__ ((__aligned__(8 * sizeof(float)))) ;
  int i;
  float f;
  double acc;

  /* 
     Compute va = va * vb + vc N times.

     va, vb and vc are initialized with values based on the first argument.
     Finally, print the sum of the elements of va, just to avoid that the optimiser optimise away the code.
  */

  if (argc != 2)
    {
      fprintf(stderr, "Usage:vectors <string>, where string is used as a seed for the computation\n");
      return -1;
    }

  f = (((float)*argv[1]) * 3.)/111111111.;

  for (i = 0; i < VECT_SIZE; i++)
    {
      va[i] = f;
      f += (1. / 11111111.);
    }


  for (i = 0; i < VECT_SIZE; i++)
    {
      vb[i] = 1. + f / 10000.;
      f += (1. / 1111111.);
    }


  for (i = 0; i < VECT_SIZE; i++)
    {
      vc[i] = f;
      f += (1. / 11111111.);
    }


  for (i = 0; i < ITERATIONS; i++)
    {
      mul3_vec(va, vb, temp, VECT_SIZE);
      add3_vec(temp, vc, va, VECT_SIZE);
    }


  acc = 0.;

  for (i = 0; i < VECT_SIZE; i++)
    acc += va[i];

  printf("Result %f\n", acc);


  return 0;
}
